#!/usr/bin/env python3
# -----------------------------------------------------------------------------
# Created on : 21.04.2022
# by : jzellinger
# SVN  $
#
#
# --- imports -----------------------------------------------------------------
import json
import os
from PIL import Image
from io import BytesIO
from minio import Minio
import numpy as np
from pycocotools.coco import COCO

class S3Helper:
    def __init__(self, **params):
        """
        Creates a new S3Helper instance.
        If no parameters are given, the helper tries to parse them from environment variables.
        Otherwise, the following parameter keys need to be used:
        -run_id: GUID of the current airflow DAG run
        -url: Adress of the MinIO server console
        -user: Username of MinIO user with read/write rights on all user buckets
        -password: Password of MinIO user with read/write rights on all user buckets

        :type params: object
        """
        if not params:
            # get params
            try:
                self.__run_id = os.environ['run_id']
                self.__url = os.environ['url']
                self.__user = os.environ['user']
                self.__password = os.environ['password']
            except Exception as e:
                raise Exception("=== unable to read {} from environment variables! ===".format(e))
        else:
            # get params
            try:
                self.__run_id = params['run_id']
                self.__url = params['url']
                self.__user = params['user']
                self.__password = params['password']
            except Exception as e:
                raise Exception("=== unable to read {} from given parameters! ===".format(e))

        self.client = Minio(self.__url,
                            access_key=self.__user,
                            secret_key=self.__password,
                            secure=False)
        self.tasks_bucket = 'tasks'

    def read_image_as_pil(self, user_id, dataset_id, image_name, target_size: (tuple or None) = None):
        """
        reads a single image based on the given parameters as a PIL object.
        :param image_name:
        :param dataset_id:
        :param user_id:
        :return:
        """
        try:
            download_object = self.client.get_object(user_id,
                                                     '{}/{}'.format(dataset_id, image_name))
            image = Image.open(download_object)
            if target_size:
                image = image.resize(target_size)
        finally:
            download_object.close()
            download_object.release_conn()

        return image


    def read_image_as_numpy_array(self, user_id, dataset_id, image_name, target_size: (tuple or None) = None):
        """
        reads a single image based on the given parameters as a numpy array
        :param user_id:
        :param dataset_id:
        :param image_name:
        :return:
        """
        return np.array(self.read_image_as_pil(user_id, dataset_id, image_name, target_size))


    def read_pointcloud_as_numpy_array(self, user_id, dataset_id, pcd_name):
        """
        reads a single pointcloud based on the givcen parameters as a numpy array
        :param user_id:
        :param dataset_id:
        :param pcd_name:
        :return:
        """

        try:
            download_object = self.client.get_object(user_id,
                                                     '{}/{}'.format(dataset_id, pcd_name))

            pcd = np.load(BytesIO(download_object.read()), allow_pickle=True)

        finally:
            download_object.close()
            download_object.release_conn()

        return pcd


    def read_coco_annotation_from_json(self, user_id, dataset_id, annotation_name):
        """
        returns a list of labels loaded from a json file as COCO object
        :param annotation_name:
        :param user_id:
        :param dataset_id:
        :return: annotation: COCO
        """

        try:
            response = self.client.get_object(user_id,
                                                     '{}/annotations/{}'.format(dataset_id, annotation_name))

            print('loading coco annotations into memory...', end='')
            annotation_dict = json.loads(response.read())
            annotation = COCO()
            dataset = annotation_dict
            assert type(dataset) == dict, 'annotation file format {} not supported'.format(type(dataset))
            print('done.')
            annotation.dataset = dataset
            annotation.createIndex()

        finally:
            response.close()
            response.release_conn()

        return annotation


    def read_dataset_files(self, user_id, dataset_id):
        """
        returns a list of all the files in a dataset
        :param user_id: 
        :param dataset_id: 
        :return: 
        """""

        return self.client.list_objects(user_id, dataset_id, recursive=True)


    def write_result_to_json(self, filename: str, data: dict):
        """
        writes a dictionary to tasks bucket as json file into the tasks result folder
        :param filename:
        :param data:
        :return:
        """

        data = json.dumps(data, indent=4).encode('UTF-8')

        self.client.put_object(self.tasks_bucket, "{}/result/{}.json".format(self.__run_id, filename),
                               BytesIO(bytes(data)), len(data))

    def store_result_to_file(self, filename: str, data: BytesIO):
        """
        stores a file represented as BytesIO to the task results
        :param filename:
        :param data:
        :return:
        """
        # setting streampoint to end to tell the size of the object
        data.seek(0,2)
        data_size = data.tell()

        # setting streampointer back to beginning
        data.seek(0)
        self.client.put_object(self.tasks_bucket, "{}/result/{}".format(self.__run_id, filename),
                               data, data_size)



    def read_input_from_json(self, filename):
        """
        reads input data in json format into a dictionary
        :return:
        """

        try:
            response = self.client.get_object(self.tasks_bucket, "{}/input/{}.json".format(self.__run_id, filename))
            data = json.loads(response.read())
        finally:
            response.close()
            response.release_conn()

        return data

    def get_input_filelist(self):
        """

        :return:
        """

        return self.client.list_objects(self.tasks_bucket, "input", recursive=True)


    def load_torch_model(self, user_id, model_name):
        """
        Loads models or weights from .pt fiels into a binary object, which can be loaded with 
        pytorch using torch.load()
        :param user_id: 
        :param model_name: 
        :return: data: BytesIO 
        """

        try:
            response = self.client.get_object(user_id, "models/{}.pt".format(model_name))
            data = BytesIO(response.read())
        finally:
            response.close()
            response.release_conn()

        return data
